use std::{
    env::current_dir,
    ffi::OsStr,
    io::{BufRead, BufReader, Write},
    net::{TcpListener, TcpStream},
    path::PathBuf,
};

static CLRF: &str = "\r\n";

fn main() {
    let listener =
        TcpListener::bind("0.0.0.0:8080").expect("Failed to create TcpListener on 0.0.0.0:8080");

    for stream in listener.incoming() {
        let mut stream = stream.unwrap(); // safe because TcpListener::incoming never returns None.

        handle_connection(stream);
    }
}

#[derive(Debug)]
enum RequestMethod {
    GET,
    POST,
}
#[derive(Debug)]
struct Request {
    method: RequestMethod,
    path: Option<String>,
}
impl Request {
    pub fn from(string: &str) -> Self {
        let mut parts = string.split(" ");

        Self {
            method: match parts.next().expect("Invalid request") {
                "GET" => RequestMethod::GET,
                "POST" => RequestMethod::POST,
                _ => panic!("Unknown method"), // TODO: fix PUT crashing entire website
            },
            path: Some(
                parts
                    .next()
                    .expect("Invalid request")
                    .split("?")
                    .next()
                    .unwrap()
                    .to_string(),
            ),
        }
    }
    pub fn file(&self) -> Option<PathBuf> {
        if self.path == None {
            return None;
        }

        let mut path = PathBuf::new();
        path.push("public");

        path.push(
            PathBuf::from(self.path.as_ref().unwrap())
                .strip_prefix("/")
                .expect("Failed to strip prefix"),
        );

        if path.extension() == None {
            path.push("index.html");
        }

        let mut cwd = current_dir().expect("Failed to get CWD");
        cwd.push(path);
        Some(cwd)
    }
}

fn get_mime_type<'a>(extension: Option<&'a OsStr>) -> String {
    if extension.is_none() {
        return "text/text".to_string();
    }

    let s = extension.unwrap().to_str();
    match s {
        Some("js") => "text/javascript".to_string(),
        Some("md") => "text/markdown".to_string(),
        Some("css") => "text/css".to_string(),
        Some("html") => "text/html".to_string(),
        Some("png") => "image/png".to_string(),
        _ => "text/text".to_string(),
    }
}

fn handle_connection(mut stream: TcpStream) {
    let buf_reader = BufReader::new(&stream);
    let http_request: Vec<_> = buf_reader
        .lines()
        .map(|result| result.unwrap())
        .take_while(|line| !line.is_empty())
        .collect();

    println!("Request: {http_request:#?}");

    if http_request.len() == 0 {
        return;
    }

    let req = Request::from(http_request.get(0).expect("Empty request"));

    println!("Requested file: {:#?}", req.file());

    let mut response_content = vec![];

    let response = match req.file() {
        Some(path) => match std::fs::read(&path) {
            Ok(content) => {
                let len = &content.len();
                response_content = content;
                format!(
                "HTTP/1.1 200 OK{0}Content-Type: {2}{0}Content-Length: {1}{0}{0}",
                CLRF,
                len,
                get_mime_type(path.extension())
            )},
            Err(e) => format!("HTTP/1.1 404{0}{0}{e}", CLRF),
        },
        None => format!("HTTP/1.1 404{0}{0}", CLRF),
    };

    stream.write_all(response.as_bytes()).unwrap();
    stream.write_all(&response_content).unwrap();
}
